import multiprocessing
from servicelayer import env

# Redis cache
# URL format: redis://localhost:6379/0
REDIS_URL = env.get("REDIS_URL")
REDIS_SHORT = 84700
REDIS_LONG = REDIS_SHORT * 200
REDIS_EXPIRE = env.to_int("REDIS_EXPIRE", REDIS_SHORT * 7)
REDIS_PREFIX = "sla"

# Persistent database tags
TAGS_DATABASE_URI = env.get("TAGS_DATABASE_URI", "sqlite://")

# Worker
WORKER_RETRY = env.to_int("WORKER_RETRY", 3)
WORKER_THREADS = env.to_int("WORKER_THREADS", multiprocessing.cpu_count())
WORKER_REPORTING = env.to_bool("WORKER_REPORTING", True)

# Amazon client credentials
AWS_KEY_ID = env.get("AWS_ACCESS_KEY_ID")
AWS_SECRET_KEY = env.get("AWS_SECRET_ACCESS_KEY")
AWS_REGION = env.get("AWS_REGION", "eu-west-1")
# S3 compatible Minio host if using Minio for storage
ARCHIVE_ENDPOINT_URL = env.get("ARCHIVE_ENDPOINT_URL")

# Storage type (either 's3', 'gs', or 'file', i.e. local file system):
ARCHIVE_TYPE = env.get("ARCHIVE_TYPE", "file")
ARCHIVE_BUCKET = env.get("ARCHIVE_BUCKET")
ARCHIVE_PATH = env.get("ARCHIVE_PATH")
PUBLICATION_BUCKET = env.get("PUBLICATION_BUCKET", ARCHIVE_BUCKET)

# Logging
LOG_FORMAT = env.get("LOG_FORMAT", "TEXT")  # options are: TEXT or JSON

# Task queue
RABBITMQ_URL = env.get("RABBITMQ_URL", "rabbitmq")
RABBITMQ_USERNAME = env.get("RABBITMQ_USERNAME", "guest")
RABBITMQ_PASSWORD = env.get("RABBITMQ_PASSWORD", "guest")
RABBITMQ_HEARTBEAT = env.to_int("RABBITMQ_HEARTBEAT", 600)
RABBITMQ_BLOCKED_CONNECTION_TIMEOUT = env.to_int(
    "RABBITMQ_BLOCKED_CONNECTION_TIMEOUT", 300
)
QUEUE_ALEPH = "aleph_queue"
QUEUE_INGEST = "ingest_queue"
QUEUE_INDEX = "index_queue"



from unittest import TestCase
from unittest.mock import patch
import json


from servicelayer import settings
from servicelayer.cache import get_fakeredis
from servicelayer.taskqueue import (
    Worker,
    Dataset,
    Task,
    get_rabbitmq_connection,
    dataset_from_collection_id,
)


class CountingWorker(Worker):
    def dispatch_task(self, task):
        assert isinstance(task, Task), task
        if not hasattr(self, "test_done"):
            self.test_done = 0
        self.test_done += 1
        self.test_task = task


class TaskQueueTest(TestCase):
    def test_task_queue(self):
        settings.QUEUE_INGEST = "sls-queue-ingest"
        conn = get_fakeredis()
        collection_id = 2
        task_id = "test-task"
        body = {
            "collection_id": 2,
            "job_id": "test-job",
            "task_id": "test-task",
            "operation": "test-op",
            "context": {},
            "payload": {},
        }
        connection = get_rabbitmq_connection()
        channel = connection.channel()
        channel.queue_purge(settings.QUEUE_INGEST)
        channel.basic_publish(
            exchange="",
            routing_key=settings.QUEUE_INGEST,
            body=json.dumps(body),
        )
        dataset = Dataset(conn=conn, name=dataset_from_collection_id(collection_id))
        dataset.add_task(task_id)
        channel.close()

        status = dataset.get_status()
        assert status["finished"] == 0, status
        assert status["pending"] == 1, status
        assert status["running"] == 0, status

        worker = CountingWorker(
            queues=[settings.QUEUE_INGEST], conn=conn, num_threads=1
        )
        worker.process(blocking=False)

        status = dataset.get_status()
        assert status["finished"] == 0, status
        assert status["pending"] == 0, status
        assert status["running"] == 1, status
        assert worker.test_done == 1
        task = Task(**body, delivery_tag=None)
        assert task.get_retry_count(conn) == 1

        with patch("servicelayer.settings.WORKER_RETRY", 0):
            channel = connection.channel()
            channel.queue_purge(settings.QUEUE_INGEST)
            channel.basic_publish(
                exchange="",
                routing_key=settings.QUEUE_INGEST,
                body=json.dumps(body),
            )
            dataset = Dataset(conn=conn, name=dataset_from_collection_id(collection_id))
            dataset.add_task(task_id)
            channel.close()
            with self.assertLogs(level="ERROR") as ctx:
                worker.process(blocking=False)
            assert "Max retries reached for task test-task. Aborting." in ctx.output[0]
            # Assert that retry count stays the same
            assert task.get_retry_count(conn) == 1

        worker.ack_message(worker.test_task, channel)
        status = dataset.get_status()
        assert status["finished"] == 1, status
        assert status["pending"] == 0, status
        assert status["running"] == 0, status



import logging
from fakeredis import FakeRedis
from redis import ConnectionPool, Redis, BusyLoadingError

from servicelayer import settings
from servicelayer.util import service_retries, backoff

log = logging.getLogger(__name__)


def get_fakeredis():
    if not hasattr(settings, "_redis_fake"):
        settings._redis_fake = FakeRedis(decode_responses=True)
    return settings._redis_fake


def get_redis_pool():
    if not hasattr(settings, "_redis_pool"):
        pool = ConnectionPool.from_url(settings.REDIS_URL, decode_responses=True)
        settings._redis_pool = pool
        wait_for_redis(pool)
    return settings._redis_pool


def get_redis():
    """Create a redis connection."""
    if settings.REDIS_URL is None:
        return get_fakeredis()
    conn = Redis(connection_pool=get_redis_pool(), decode_responses=True)
    return conn


def wait_for_redis(pool):
    """Wait for redis to load its data into memory on initial system
    bootup."""
    for attempt in service_retries():
        try:
            conn = Redis(connection_pool=pool, decode_responses=True)
            conn.ping()
            return
        except BusyLoadingError:
            log.info("Waiting for redis to load...")
            backoff(failures=attempt)
    raise RuntimeError("Redis is not ready.")


def make_key(*criteria):
    """Make a string key out of many criteria."""
    parts = []
    for criterion in criteria:
        if criterion is None:
            continue
        criterion = str(criterion)
        criterion = criterion.replace(":", "#")
        parts.append(criterion)
    return ":".join(parts)



from dataclasses import dataclass
from typing import Optional
import json
import time
import threading
import signal
import logging
import sys
from abc import ABC, abstractmethod
import functools
from queue import Queue, Empty

from structlog.contextvars import clear_contextvars, bind_contextvars
import pika
from banal import ensure_list

from servicelayer.cache import get_redis, make_key
from servicelayer.util import unpack_int
from servicelayer import settings
from servicelayer.util import service_retries, backoff

log = logging.getLogger(__name__)
local = threading.local()


PREFIX = "tq"
NO_COLLECTION = "null"


OP_INGEST = "ingest"
OP_ANALYZE = "analyze"
OP_INDEX = "index"

# ToDo: consider ALEPH_INGEST_PIPELINE setting
INGEST_OPS = (OP_INGEST, OP_ANALYZE)

TIMEOUT = 5


@dataclass
class Task:
    task_id: str
    job_id: str
    delivery_tag: str
    operation: str
    context: dict
    payload: dict
    collection_id: Optional[str] = None

    @property
    def retry_key(self):
        dataset = dataset_from_collection_id(self.collection_id)
        return make_key(PREFIX, "qdj", dataset, "taskretry", self.task_id)

    def get_retry_count(self, conn):
        return unpack_int(conn.get(self.retry_key))

    def increment_retry_count(self, conn):
        conn.incr(self.retry_key)
        conn.expire(self.retry_key, settings.REDIS_EXPIRE)

    def get_dataset(self, conn):
        dataset = Dataset(
            conn=conn, name=dataset_from_collection_id(self.collection_id)
        )
        return dataset


class Dataset:
    """Track the status of tasks for a given dataset/collection"""

    def __init__(self, conn, name):
        self.conn = conn
        self.name = name
        self.key = make_key(PREFIX, "qdatasets")
        # integer that tracks number of finished tasks
        self.finished_key = make_key(PREFIX, "qdj", name, "finished")
        # sets that contain task ids of running and pending tasks
        self.running_key = make_key(PREFIX, "qdj", name, "running")
        self.pending_key = make_key(PREFIX, "qdj", name, "pending")

    def cancel(self):
        """Cancel processing of all tasks belonging to a dataset"""
        pipe = self.conn.pipeline()
        # remove the dataset from active datasets
        pipe.srem(self.key, self.name)
        # clean up tasks and task counts
        pipe.delete(self.finished_key)
        pipe.delete(self.running_key)
        pipe.delete(self.pending_key)
        pipe.execute()

    def get_status(self):
        """Status of a given dataset."""
        status = {"finished": 0, "running": 0, "pending": 0}
        finished = self.conn.get(self.finished_key)
        running = self.conn.scard(self.running_key)
        pending = self.conn.scard(self.pending_key)
        status["finished"] = max(0, unpack_int(finished))
        status["running"] = max(0, unpack_int(running))
        status["pending"] = max(0, unpack_int(pending))
        return status

    @classmethod
    def get_active_dataset_status(cls, conn):
        """Status of all active datasets."""
        result = {"total": 0, "datasets": {}}
        datasets_key = make_key(PREFIX, "qdatasets")
        for name in conn.smembers(datasets_key):
            dataset = cls(conn, name)
            status = dataset.get_status()
            result["total"] += 1
            result["datasets"][dataset.name] = status
        return result

    @classmethod
    def cleanup_dataset_status(cls, conn):
        """Clean up dataset status for inactive datasets."""
        datasets_key = make_key(PREFIX, "qdatasets")
        for name in conn.smembers(datasets_key):
            dataset = cls(conn, name)
            status = dataset.get_status()
            if status["running"] == 0 and status["pending"] == 0:
                pipe = conn.pipeline()
                # remove the dataset from active datasets
                pipe.srem(dataset.key, dataset.name)
                # reset finished task count
                pipe.delete(dataset.finished_key)
                pipe.execute()

    def should_execute(self, task_id):
        """Should a task be executed?

        When a the processing of a task is cancelled, there is no way to tell RabbitMQ to drop it.
        So we store that state in Redis and make the worker check with Redis before executing a task.
        """
        attempt = 1
        while True:
            _should_execute = self.conn.sismember(
                self.pending_key, task_id
            ) or self.conn.sismember(self.running_key, task_id)
            if not _should_execute and attempt < settings.WORKER_RETRY:
                # Sometimes for new tasks the check fails because the Redis
                # state gets updated only after the task gets written to disk
                # by RabbitMQ whereas the worker consumer gets the task before
                # that.
                # Retry a few times to avoid those cases.
                backoff(failures=attempt)
                attempt += 1
                continue
            return _should_execute

    def add_task(self, task_id):
        """Update state when a new task is added to the task queue"""
        log.info(f"Adding task: {task_id}")
        pipe = self.conn.pipeline()
        # add the dataset to active datasets
        pipe.sadd(self.key, self.name)
        pipe.sadd(self.pending_key, task_id)
        pipe.execute()

    def remove_task(self, task_id):
        """Remove a task that's not going to be executed"""
        log.info(f"Removing task: {task_id}")
        self.conn.srem(self.pending_key, task_id)
        status = self.get_status()
        if status["running"] == 0 and status["pending"] == 0:
            # remove the dataset from active datasets
            self.conn.srem(self.key, self.name)

    def checkout_task(self, task_id):
        """Update state when a task is checked out for execution"""
        log.info(f"Checking out task: {task_id}")
        pipe = self.conn.pipeline()
        # add the dataset to active datasets
        pipe.sadd(self.key, self.name)
        pipe.srem(self.pending_key, task_id)
        pipe.sadd(self.running_key, task_id)
        pipe.execute()

    def mark_done(self, task: Task):
        """Update state when a task is finished executing"""
        log.info(f"Finished executing task: {task.task_id}")
        pipe = self.conn.pipeline()
        pipe.srem(self.pending_key, task.task_id)
        pipe.srem(self.running_key, task.task_id)
        pipe.incr(self.finished_key)
        pipe.delete(task.retry_key)
        pipe.execute()
        status = self.get_status()
        if status["running"] == 0 and status["pending"] == 0:
            # remove the dataset from active datasets
            self.conn.srem(self.key, self.name)

    def is_done(self):
        status = self.get_status()
        return status["pending"] == 0 and status["running"] == 0

    def __str__(self):
        return self.name


def get_task(body, delivery_tag) -> Task:
    body = json.loads(body)
    return Task(
        collection_id=collection_id_from_dataset(body["collection_id"]),
        task_id=body["task_id"],
        job_id=body["job_id"],
        delivery_tag=delivery_tag,
        operation=body["operation"],
        context=body["context"] or {},
        payload=body["payload"] or {},
    )


def dataset_from_collection_id(collection_id):
    """servicelayer dataset from a collection"""
    if collection_id is None:
        return NO_COLLECTION
    return str(collection_id)


def collection_id_from_dataset(dataset):
    """Invert the servicelayer dataset into a collection ID"""
    if dataset == NO_COLLECTION:
        return None
    return int(dataset)


def apply_task_context(task: Task, **kwargs):
    """This clears the current structured logging context and readies it
    for a new task"""
    # Setup context for structured logging
    clear_contextvars()
    bind_contextvars(
        job_id=task.job_id,
        stage=task.operation,
        dataset=dataset_from_collection_id(task.collection_id),
        start_time=time.time(),
        trace_id=task.task_id,
        **kwargs,
    )


def get_routing_key(stage):
    if stage in INGEST_OPS:
        routing_key = settings.QUEUE_INGEST
    elif stage == OP_INDEX:
        routing_key = settings.QUEUE_INDEX
    else:
        routing_key = settings.QUEUE_ALEPH
    return routing_key


class MaxRetriesExceededError(Exception):
    pass


class Worker(ABC):
    def __init__(
        self,
        queues,
        conn=None,
        num_threads=settings.WORKER_THREADS,
        version=None,
        prefetch_count=100,
    ):
        self.conn = conn or get_redis()
        self.num_threads = num_threads
        self.queues = ensure_list(queues)
        self.version = version
        self.prefetch_count = prefetch_count
        self.local_queue = Queue()

    def on_signal(self, signal, _):
        log.warning(f"Shutting down worker (signal {signal})")
        # Exit eagerly without waiting for current task to finish running
        sys.exit(int(signal))

    def on_message(self, channel, method, properties, body, args):
        """RabbitMQ on_message event handler.

        We have to make sure it doesn't block for long to ensure that RabbitMQ
        heartbeats are not interrupted.
        """
        connection = args[0]
        task = get_task(body, method.delivery_tag)
        self.local_queue.put((task, channel, connection))

    def process_blocking(self):
        """Blocking worker thread - executes tasks from a queue and periodic tasks"""
        while True:
            try:
                (task, channel, connection) = self.local_queue.get(timeout=TIMEOUT)
                apply_task_context(task, v=self.version)
                self.handle(task)
                cb = functools.partial(self.ack_message, task, channel)
                connection.add_callback_threadsafe(cb)
            except Empty:
                pass
            finally:
                clear_contextvars()
                self.periodic()

    def process_nonblocking(self):
        """Non-blocking worker is used for tests only."""
        connection = get_rabbitmq_connection()
        channel = connection.channel()
        queue_active = {queue: True for queue in self.queues}
        while True:
            for queue in self.queues:
                method, properties, body = channel.basic_get(queue=queue)
                if method is None:
                    queue_active[queue] = False
                    # Quit processing if all queues are inactive
                    if not any(queue_active.values()):
                        return
                else:
                    queue_active[queue] = True
                    task = get_task(body, method.delivery_tag)
                    self.handle(task)

    def process(self, blocking=True):
        if blocking:
            self.process_blocking()
        else:
            self.process_nonblocking()

    def handle(self, task: Task):
        """Execute a task."""
        try:
            dataset = Dataset(
                conn=self.conn, name=dataset_from_collection_id(task.collection_id)
            )
            if dataset.should_execute(task.task_id):
                if task.get_retry_count(self.conn) > settings.WORKER_RETRY:
                    raise MaxRetriesExceededError(
                        f"Max retries reached for task {task.task_id}. Aborting."
                    )
                dataset.checkout_task(task.task_id)
                task.increment_retry_count(self.conn)
                task = self.dispatch_task(task)
            else:
                log.warn(f"Discarding task: {task.task_id}")
        except Exception:
            log.exception("Error in task handling")
        finally:
            self.after_task(task)

    @abstractmethod
    def dispatch_task(self, task: Task) -> Task:
        raise NotImplementedError

    def after_task(self, task: Task):
        """Run after-task clean up"""
        pass

    def periodic(self):
        """Periodic tasks to run."""
        pass

    def ack_message(self, task, channel):
        """Acknowledge a task after execution.

        RabbitMQ requires that the channel used for receiving the message must be used
        for acknowledging a message as well.
        """
        apply_task_context(task, v=self.version)
        skip_ack = task.context.get("skip_ack")
        if skip_ack:
            log.info(
                f"Skipping acknowledging message {task.delivery_tag} for task_id {task.task_id}"
            )
        else:
            log.info(
                f"Acknowledging message {task.delivery_tag} for task_id {task.task_id}"
            )
            dataset = task.get_dataset(conn=self.conn)
            # Sync state to redis
            dataset.mark_done(task)
            if channel.is_open:
                channel.basic_ack(task.delivery_tag)
        clear_contextvars()

    def run(self):
        """Run a blocking worker instance"""

        # Handle kill signals
        signal.signal(signal.SIGINT, self.on_signal)
        signal.signal(signal.SIGTERM, self.on_signal)

        # worker threads
        process = lambda: self.process(blocking=True)
        threads = []
        for _ in range(self.num_threads):
            thread = threading.Thread(target=process)
            thread.daemon = True
            thread.start()
            threads.append(thread)

        log.info(f"Worker has {self.num_threads} worker threads.")

        connection = get_rabbitmq_connection()
        channel = connection.channel()
        channel.basic_qos(prefetch_count=self.prefetch_count)
        on_message_callback = functools.partial(self.on_message, args=(connection,))
        for queue in self.queues:
            channel.queue_declare(queue=queue, durable=True)
            channel.basic_consume(queue=queue, on_message_callback=on_message_callback)
        channel.start_consuming()


def get_rabbitmq_connection():
    for attempt in service_retries():
        try:
            if not hasattr(local, "connection") or not local.connection:
                credentials = pika.PlainCredentials(
                    settings.RABBITMQ_USERNAME, settings.RABBITMQ_PASSWORD
                )
                connection = pika.BlockingConnection(
                    pika.ConnectionParameters(
                        host=settings.RABBITMQ_URL,
                        credentials=credentials,
                        heartbeat=settings.RABBITMQ_HEARTBEAT,
                        blocked_connection_timeout=settings.RABBITMQ_BLOCKED_CONNECTION_TIMEOUT,
                    )
                )
                local.connection = connection
            if local.connection.is_open:
                channel = local.connection.channel()
                channel.queue_declare(queue=settings.QUEUE_ALEPH, durable=True)
                channel.queue_declare(queue=settings.QUEUE_INGEST, durable=True)
                channel.queue_declare(queue=settings.QUEUE_INDEX, durable=True)
                channel.close()
                return local.connection
        except (pika.exceptions.AMQPConnectionError, pika.exceptions.AMQPError):
            log.exception("RabbitMQ error")
        local.connection = None
        backoff(failures=attempt)
    raise RuntimeError("Could not connect to RabbitMQ")



from unittest import TestCase

from servicelayer.extensions import get_extensions


class ExtensionTest(TestCase):
    def test_extensions(self):
        exts = get_extensions("servicelayer.test")
        assert len(exts), exts
        assert get_extensions in exts, exts



from unittest import TestCase
from datetime import datetime

from servicelayer.util import pack_datetime, pack_now, unpack_datetime
from servicelayer.util import pack_int, unpack_int
from servicelayer.util import dump_json, load_json


class UtilTest(TestCase):
    def test_int(self):
        packed = pack_int(555)
        unpacked = unpack_int(packed)
        assert unpacked == 555
        assert unpack_int("Banana") == 0

    def test_datetime(self):
        date = datetime.utcnow().date()
        packed = pack_datetime(date)
        unpacked = unpack_datetime(packed)
        assert unpacked == date
        now = pack_now()
        assert now.startswith(packed), now
        time = datetime.utcnow()
        packed = pack_datetime(time)
        unpacked = unpack_datetime(packed)
        assert unpacked == time
        assert pack_datetime("Banana") is None
        assert unpack_datetime("Banana") is None

    def test_object(self):
        data = {
            "help": 5,
            "xxx": datetime.utcnow(),
            "bla": None,
            "foo": set(["a"]),
            "yum": b"xxx",
        }
        json = dump_json(data)
        out = load_json(json)
        assert "bla" not in out, out
        assert "xxx" in out, out
        assert "a" in out["foo"], out
        assert load_json("") is None
        assert load_json(None) is None
        assert dump_json(None) == ""



from unittest import TestCase
import pytest

from servicelayer.cache import get_fakeredis
from servicelayer.jobs import Job
from servicelayer import worker


worker.TASK_FETCH_RETRY = 1


class CountingWorker(worker.Worker):
    def boot(self):
        self.test_done = 0

    def handle(self, task):
        self.test_done += 1


class WorkerTest(TestCase):
    def test_run(self):
        conn = get_fakeredis()
        operation = "lala"
        worker = CountingWorker(conn=conn, stages=[operation])
        worker.sync()
        assert worker.test_done == 0, worker.test_done
        job = Job.create(conn, "test")
        stage = job.get_stage(operation)
        task = stage.queue({}, {})
        assert not job.is_done()
        assert worker.test_done == 0, worker.test_done
        worker.sync()
        assert worker.test_done == 1, worker.test_done
        assert job.is_done()
        worker.retry(task)
        assert not job.is_done()
        worker.sync()
        assert job.is_done()
        assert worker.exit_code == 0, worker.exit_code
        assert worker.test_done == 1, worker.test_done
        worker.retry(task)
        worker.run(blocking=False)
        assert job.is_done()
        assert worker.exit_code == 0, worker.exit_code
        worker.num_threads = None
        worker.retry(task)
        worker.run(blocking=False)
        assert job.is_done()
        assert worker.exit_code == 0, worker.exit_code
        try:
            worker._handle_signal(5, None)
        except SystemExit as exc:
            assert exc.code == 5, exc.code
        with pytest.raises(SystemExit) as exc:  # noqa
            worker._handle_signal(5, None)



import boto3
import logging
from datetime import datetime
from botocore.exceptions import ClientError

from servicelayer import settings
from servicelayer.archive.virtual import VirtualArchive
from servicelayer.archive.util import checksum, ensure_path
from servicelayer.archive.util import path_prefix, path_content_hash

log = logging.getLogger(__name__)


class S3Archive(VirtualArchive):
    TIMEOUT = 84600

    def __init__(self, bucket=None, publication_bucket=None):
        super(S3Archive, self).__init__(bucket)
        key_id = settings.AWS_KEY_ID
        secret_key = settings.AWS_SECRET_KEY
        self.client = boto3.client(
            "s3",
            endpoint_url=settings.ARCHIVE_ENDPOINT_URL,
            region_name=settings.AWS_REGION,
            aws_access_key_id=key_id,
            aws_secret_access_key=secret_key,
        )
        # config=Config(signature_version='s3v4'))
        self.bucket = bucket
        self.publication_bucket = publication_bucket
        log.info("Archive: s3://%s", bucket)

        try:
            self.client.head_bucket(Bucket=bucket)
        except ClientError as e:
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                self.client.create_bucket(
                    Bucket=bucket,
                    CreateBucketConfiguration={
                        "LocationConstraint": settings.AWS_REGION
                    },
                )
                self.upgrade()
            else:
                log.exception("Could not check bucket")

    def upgrade(self):
        # Make sure bucket policy is set correctly.
        config = {
            "CORSRules": [
                {
                    "AllowedMethods": ["GET"],
                    "AllowedOrigins": ["*"],
                    "AllowedHeaders": ["*"],
                    "ExposeHeaders": [
                        "Accept-Ranges",
                        "Content-Encoding",
                        "Content-Length",
                        "Content-Range",
                    ],
                    "MaxAgeSeconds": self.TIMEOUT,
                }
            ]
        }
        try:
            self.client.put_bucket_cors(Bucket=self.bucket, CORSConfiguration=config)
        except ClientError:
            log.exception("Could not update CORS")

    def _locate_key(self, content_hash=None, prefix=None):
        """Check if a file with the given hash exists on S3."""
        if prefix is None:
            if content_hash is None:
                return
            prefix = path_prefix(content_hash)
            if prefix is None:
                return
        res = self.client.list_objects(MaxKeys=1, Bucket=self.bucket, Prefix=prefix)
        for obj in res.get("Contents", []):
            return obj.get("Key")

    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Store the file located at the given path on S3, based on a path
        made up from its SHA1 content hash."""
        file_path = ensure_path(file_path)
        if content_hash is None:
            content_hash = checksum(file_path)

        # if content_hash is None:
        #     return

        obj = self._locate_key(content_hash)
        if obj is not None:
            return content_hash

        path = "{}/data".format(path_prefix(content_hash))
        extra = {}
        if mime_type is not None:
            extra["ContentType"] = mime_type
        with open(file_path, "rb") as fh:
            self.client.upload_fileobj(fh, self.bucket, str(path), ExtraArgs=extra)
        return content_hash

    def load_file(self, content_hash, file_name=None, temp_path=None):
        """Retrieve a file from S3 storage and put it onto the local file
        system for further processing."""
        key = self._locate_key(content_hash)
        if key is not None:
            path = self._local_path(content_hash, file_name, temp_path)
            self.client.download_file(self.bucket, key, str(path))
            return path

    def delete_file(self, content_hash):
        if content_hash is None:
            return
        prefix = path_prefix(content_hash)
        if prefix is None:
            return
        res = self.client.list_objects(Bucket=self.bucket, Prefix=prefix)
        for obj in res.get("Contents", []):
            self.client.delete_object(Bucket=self.bucket, Key=obj.get("Key"))

    def list_files(self, prefix=None):
        """Try to list out all the hashes in the archive."""
        kwargs = {"Bucket": self.bucket}
        prefix = path_prefix(prefix)
        if prefix is not None:
            kwargs["Prefix"] = prefix
        token = None
        while True:
            if token is not None:
                kwargs["ContinuationToken"] = token
            res = self.client.list_objects_v2(**kwargs)
            for obj in res.get("Contents", []):
                yield path_content_hash(obj.get("Key"))
            if not res.get("IsTruncated"):
                break
            token = res.get("NextContinuationToken")

    def generate_url(self, content_hash, file_name=None, mime_type=None, expire=None):
        key = self._locate_key(content_hash)
        if key is None:
            return
        params = {"Bucket": self.bucket, "Key": key}
        if mime_type is not None:
            params["ResponseContentType"] = mime_type
        if file_name is not None:
            disposition = "inline; filename=%s" % file_name
            params["ResponseContentDisposition"] = disposition
        expires_in = self.TIMEOUT
        if expire is not None:
            delta = expire - datetime.utcnow()
            expires_in = int(delta.total_seconds())
        return self.client.generate_presigned_url(
            "get_object", Params=params, ExpiresIn=expires_in
        )

    @property
    def can_publish(self):
        return True

    def publish_file(self, file_path, publish_path, mime_type=None):
        extra = {"ACL": "public-read"}
        if mime_type is not None:
            extra["ContentType"] = mime_type
        with open(file_path, "rb") as fh:
            self.client.upload_fileobj(
                fh, self.publication_bucket, publish_path, ExtraArgs=extra
            )
        params = {"Bucket": self.publication_bucket, "Key": publish_path}
        if mime_type is not None:
            params["ResponseContentType"] = mime_type
        return self.client.generate_presigned_url(
            "get_object", Params=params, ExpiresIn=0
        )
